/*
  name:           playfair.c
  
  syntax:         playfair [- d | p] [-i] "key string" < infile > outfile

  description:    Encipher or decipher text files using the playfair cipher
                  "-d" deciphers text; "-p" prints the playfair cipher table
                  for the key phrase. "-t" transposes the playfair key table.

  Author:         Charles Shapiro 23 November 1988

  This source code is truly public domain; you may use it for fun or profit, 
  sell it for whatever the market will bear, or make any modifications to 
  it you wish.  I ask as a matter of courtesy that you credit me with writing 
  it if you plan to distribute it or a modification of it.

*/
#include <stdio.h>
#include <ctype.h>


#define ROW1 0
#define COL1 1
#define ROW2 2
#define COL2 3

char codetbl[5][5];

static char alphabet[] = {"abcdefghiklmnopqrstuvwxyz"};

int print_sw;
int decode_sw;
int transpose_sw;

/*
This is where we check that we've got valid key and option arguments,
and set whatever options need to be set.
*/
int args_ok(arc,arv,keyptr)
  int arc;
  char *arv[];
  char **keyptr;
{
  int i;

  if((arc > 4) || (arc < 2))
    return 0;


  *keyptr = 0; 
  print_sw = 0;
  decode_sw = 0;
  transpose_sw = 0;

  for(i=1;i<arc;i++) {
    if( (*(arv[i])) == '-') {
      switch(tolower(arv[i][1])) {
        case 'd':
          decode_sw = 1;
          if(strlen(arv[i]) > 2) {
            if(tolower(arv[i][2]) == 't')
              transpose_sw = 1;
            else {
              fprintf(stderr,"\nBad option argument - %c",arv[i][2]);
              exit(1);
            }
          }
          break;
        case 'p':
          print_sw = 1;
          if(strlen(arv[i]) > 2) {
            if(tolower(arv[i][2]) == 't')
              transpose_sw = 1;
            else {
              fprintf(stderr,"\nBad option argument - %c",arv[i][2]);
              exit(1);
            }
          }
          break;
        case 't':
           transpose_sw = 1;
           break;
        default:
          printf("Bad option argument - %c\n",arv[i][1]);
          return(0);
      }
    } 
    else 
     *keyptr = arv[i];
  }
  return(-1);
}

/*
Remove spaces, duplicate letters from key.
*/
void purify_key(inbuf,outbuf)
  char *inbuf; 
  char *outbuf;
{
  int i,j,dup;
  int in_len;
  char *out_ptr;

  in_len = strlen(inbuf);
  out_ptr = outbuf;
  if(isalpha(*inbuf)) {
    *out_ptr = tolower(*inbuf);
    ++out_ptr; 
  }
  
  for(i=1;i<=in_len;i++) {
    if(! isalpha(inbuf[i]))
      continue;
    dup = 0;
    for(j=0;j<i;j++) 
      if(tolower(inbuf[j]) == tolower(inbuf[i]))
        dup=1;
    if(! dup) {
      *out_ptr = tolower(inbuf[i]);
      if(*out_ptr == 'j')
        *out_ptr='i';
      ++out_ptr;
    }
  }
  *out_ptr=0;
}

/*
Make the playfair key table.
*/
void make_keytbl(codekey)
  char *codekey;
{
  char *tablptr;
  int i;

  strncpy((char *)codetbl,codekey,25);

  tablptr = (char *)codetbl + strlen(codekey);

  for(i=0;i<25;i++) {
    if(! strchr((char *)codetbl,alphabet[i])) {
      *tablptr = alphabet[i];
      ++tablptr;
    } 
  }
}

/*
Print out the key table for manual encipher/decipher.
*/
void print_keytbl()
{
  int i;

  for(i=0;i<25;i++) {
    printf("%c ",((char *)codetbl)[i]);
    if( ! ((i+1) % 5))
      putchar('\n');
  }
}
/*
Get a letter from stdin.  Make it lowercase.  If it's a j, make it an i.
*/
int get_letter()
{
  int c;

  c = getchar();

  while((! isalpha(c)) && (c != EOF))
    c = getchar();

  c=tolower(c);
  if(c == 'j')
    c = 'i';
  return(c);
}
/*
Get next pair of letters from stdin, substituting nulls (X's) where
appropriate.  
*/
int get_pair(in_pair)
  char *in_pair;
{
  static int in_chars[2] = {0,2};
  int retval=0;

  if(in_chars[0] == EOF)
    goto exit_point; 
  

  if(in_chars[0] != in_chars[1]) {
    in_chars[0] = get_letter();
    if(in_chars[0] == EOF)
      goto exit_point;
    else
      in_chars[1] = get_letter(); 
  }
  else 
    in_chars[1] = get_letter();

  if(in_chars[1] == EOF) {
    *in_pair = (char)*in_chars;
    in_pair[1] = 0;
    in_chars[0] = EOF;
    retval = 1;
    goto exit_point;
  }
     
  if(in_chars[0] == in_chars[1]) {
    *in_pair = (char)*in_chars;
    in_pair[1] = 0;
    retval = 1;
    goto exit_point;
  }

  in_pair[0] = (char)in_chars[0];
  in_pair[1] = (char)in_chars[1];
  retval = 1;

  exit_point:
    if(retval) {
      if(! (in_pair[1])) {
        if(in_pair[0] != 'x')
          in_pair[1] = 'x';
        else
          in_pair[1] = 'y';
      }
    }
    return(retval);
}

/*
Find a letter in the key grid.
*/
void find_letter(letter,row,col)
  char letter;
  int *row;
  int *col;
{
  int i,j;

  *row = EOF;
 
  for(i=0;i<5;i++) {
    for(j=0;j<5;j++) {
      if(codetbl[i][j] == letter) {
        *row = i;
        *col = j;
      }
    }
    if(*row != EOF)
      break;
  }
}

/*
Encipher a message with the playfair cipher.
*/
void encode_playf(inpair,outpair)
  char *inpair;
  char *outpair;
{

  int thecase;
  int inpair_coords[4];
  int outpair_coords[4];

  find_letter(inpair[0],&inpair_coords[ROW1],&inpair_coords[COL1]);
  find_letter(inpair[1],&inpair_coords[ROW2],&inpair_coords[COL2]);


  if(inpair_coords[ROW1] == inpair_coords[ROW2])  /* Same row. */
    thecase = 0;
  else
    if(inpair_coords[COL1] == inpair_coords[COL2])  /* Same column */
      thecase = 1; 
  else
    thecase = 2;                                     /* same neither */
  
  switch(thecase) {
    case 0:				            /* same row. */
      outpair_coords[ROW1] = inpair_coords[ROW1];
      outpair_coords[ROW2] = inpair_coords[ROW2];
      if(inpair_coords[COL1] == 4)
        outpair_coords[COL1] = 0;
      else
        outpair_coords[COL1] = inpair_coords[COL1] + 1;
      if(inpair_coords[COL2] == 4)
        outpair_coords[COL2] = 0;
      else
        outpair_coords[COL2] = inpair_coords[COL2] + 1;
      break;
    case 1:
      outpair_coords[COL1] = inpair_coords[COL1];
      outpair_coords[COL2] = inpair_coords[COL2];
      if(inpair_coords[ROW1] == 4)
        outpair_coords[ROW1] = 0;
      else
        outpair_coords[ROW1] = inpair_coords[ROW1] + 1;
      if(inpair_coords[ROW2] == 4)
        outpair_coords[ROW2] = 0;
      else
        outpair_coords[ROW2] = inpair_coords[ROW2] + 1;
      break;
    case 2:
      outpair_coords[ROW1] = inpair_coords[ROW1];
      outpair_coords[COL1] = inpair_coords[COL2];
      outpair_coords[ROW2] = inpair_coords[ROW2];
      outpair_coords[COL2] = inpair_coords[COL1];
      break;
  }
  outpair[0] = codetbl[outpair_coords[ROW1]][outpair_coords[COL1]];
  outpair[1] = codetbl[outpair_coords[ROW2]][outpair_coords[COL2]];

}
/*
Decipher a message with the playfair cipher.
*/
void decode_playf(inpair,outpair)
  char *inpair;
  char *outpair;
{
  int thecase;
  int inpair_coords[4];
  int outpair_coords[4];

  find_letter(inpair[0],&inpair_coords[ROW1],&inpair_coords[COL1]);
  find_letter(inpair[1],&inpair_coords[ROW2],&inpair_coords[COL2]);


  if(inpair_coords[ROW1] == inpair_coords[ROW2])  /* Same row. */
    thecase = 0;
  else
    if(inpair_coords[COL1] == inpair_coords[COL2])  /* Same column */
      thecase = 1; 
  else
    thecase = 2;                                     /* same neither */
  
  switch(thecase) {
    case 0:				            /* same row. */
      outpair_coords[ROW1] = inpair_coords[ROW1];
      outpair_coords[ROW2] = inpair_coords[ROW2];
      if( ! inpair_coords[COL1])
        outpair_coords[COL1] = 4;
      else
        outpair_coords[COL1] = inpair_coords[COL1] - 1;
      if(! inpair_coords[COL2])
        outpair_coords[COL2] = 4;
      else
        outpair_coords[COL2] = inpair_coords[COL2] - 1;
      break;
    case 1:
      outpair_coords[COL1] = inpair_coords[COL1];
      outpair_coords[COL2] = inpair_coords[COL2];
      if(! inpair_coords[ROW1])
        outpair_coords[ROW1] = 4;
      else
        outpair_coords[ROW1] = inpair_coords[ROW1] - 1;
      if(! inpair_coords[ROW2])
        outpair_coords[ROW2] = 4;
      else
        outpair_coords[ROW2] = inpair_coords[ROW2] - 1;
      break;
    case 2:
      outpair_coords[ROW1] = inpair_coords[ROW1];
      outpair_coords[COL1] = inpair_coords[COL2];
      outpair_coords[ROW2] = inpair_coords[ROW2];
      outpair_coords[COL2] = inpair_coords[COL1];
      break;
  }
  outpair[0] = codetbl[outpair_coords[ROW1]][outpair_coords[COL1]];
  outpair[1] = codetbl[outpair_coords[ROW2]][outpair_coords[COL2]];

}
/*
Invert a key.
*/
char *transpose_key(key,inbuf)
  char *key;      /* the original key. */
  char *inbuf;    /* The original key with the rest of the alphabet. */
{
  int keylen = strlen(key);
  char *inbufptr;
  char *outbufptr;
  int i;
  static char transposed[128];
  char key_buf[128];


  memset(key_buf,0,sizeof(key_buf));
  strncpy(key_buf,inbuf,25);
  outbufptr = transposed;

  for(i=0;i<keylen;i++) {
    inbufptr = &(key_buf[i]);
    while(*inbufptr) {
      *outbufptr = *inbufptr;
      ++outbufptr;
      inbufptr += keylen;
    }
  }
  return(transposed);
}
/*
Main line.  
*/
main(argc,argv)
  int argc;
  char *argv[];
{

  char working_key[255];
  char *key;
  char input_pair[3];
  char output_pair[3];
  void (*process)();
  int output_counter;

  if(! args_ok(argc,argv,&key)) {
    fprintf(stderr,"\nUsage: playfair [-[d | p]] [-i] \"key string\"");
    exit(1);
  }

  purify_key(key,working_key);

  make_keytbl(working_key);

  if(transpose_sw)
    strncpy(codetbl,transpose_key(working_key,(char *)&codetbl),25);

  if(print_sw) {
    print_keytbl();
    exit(0);
  }

  if(decode_sw)
    process = decode_playf;
  else
    process = encode_playf;

  output_counter = 0;
    
  while(get_pair(input_pair)) {
    (*process)(input_pair,output_pair);  
    output_pair[2]=0;
    output_counter += 3;
    if(output_counter > 78) {
      printf("%s\n",output_pair);
      output_counter = 0;
    }
    else
      printf("%s ",output_pair); 
  }
  putchar('\n');
}

